package uk.ac.bbsrc.babraham.ChipMonk.Displays.ConditionTree;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.swing.JOptionPane;

import uk.ac.bbsrc.babraham.ChipMonk.ChipMonkException;
import uk.ac.bbsrc.babraham.ChipMonk.Analysis.Statistics.PearsonCorrelation;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.DataStore;
import uk.ac.bbsrc.babraham.ChipMonk.DataTypes.Probe;

public class DistanceMatrix implements Runnable {


	private Vector nodes = new Vector();
	private Hashtable matrix = new Hashtable();
	private Hashtable dataSets = new Hashtable();
	private Vector listeners = new Vector();
	
	private DataStore [] data;
	private Probe [] probes;
	
	public void calculateTree (DataStore [] data, Probe [] probes) {
		
		// Do some sanity checks.
		if (data.length < 3) {
			JOptionPane.showMessageDialog(null,"At least three data sets are needed to construct a tree","Error",JOptionPane.ERROR_MESSAGE);
			return;
		}
		this.data = data;
		this.probes = probes;
		Thread t = new Thread(this);
		t.start();
	}
	
	public void addListener (MatrixProgressListener l) {
		listeners.add(l);
	}
	
	public void removeListener (MatrixProgressListener l) {
		listeners.removeElement(l);
	}
	
	public void run() {
		// First we need to extract out the data we're going to use
		// to calculate the distance matrix
		for (int i=0;i<data.length;i++) {
			double [] tempData = new double [probes.length];
			for (int p=0;p<probes.length;p++) {
				tempData[p] = Math.log(data[i].getRatioForProbe(probes[p]));
			}
			ConditionTreeNode tempNode = new ConditionTreeNode(data[i]);
			dataSets.put(tempNode,tempData);
			nodes.add(tempNode);
		}
		
		// Now we need to calculate the initial matrix
		Enumeration n = nodes.elements();
		while (n.hasMoreElements()) {
			ConditionTreeNode orig = (ConditionTreeNode)n.nextElement();
			Hashtable origData = new Hashtable();
			matrix.put(orig,origData);
			boolean seenSelf = false;
			Enumeration m = nodes.elements();
			while (m.hasMoreElements()) {
				ConditionTreeNode thisNode = (ConditionTreeNode)m.nextElement();
				
				if (thisNode == orig) {
					seenSelf = true;
					continue;
				}
				
				if (seenSelf) {
					// We subtract the distance from 1 so we get a range between 0 - 2 (all positive)
					// with 0 being the most closely correlated (r of +1)
					float r=0;
					try {
						r = 1 - PearsonCorrelation.calculateCorrelation((double [])dataSets.get(orig),(double [])dataSets.get(thisNode));
					} catch (ChipMonkException e) {
						e.printStackTrace();
					}
					origData.put(thisNode,new Float(r));
				}
			}
		}
		
		// Now we need to construct the tree.  We're using a neighbor joining
		// method.  See http://artedi.ebc.uu.se/course/sommar/njoin/ for a 
		// good explanation of how this works.
		
		// The joining is an iterative process which continues until we only
		// have one node left
		ConditionTreeNode lastNodeAdded = null;
		while (nodes.size() > 2) {
//			System.out.println("Nodes vector contains "+nodes.size()+" nodes");
			NodePair closestPair = findClosestPair();
			lastNodeAdded = expandTree(closestPair);
		}
		
		// Once we're down to two nodes left we manually add the root of the tree.
		
		// The distance is the one remaining distance in the table
		ConditionTreeNode outgroup = (ConditionTreeNode)nodes.elementAt(0);
		lastNodeAdded.add(outgroup,new Float(getDistance(outgroup,lastNodeAdded)));

		Enumeration l = listeners.elements();
		while (l.hasMoreElements()) {
			((MatrixProgressListener)l.nextElement()).matrixProcessingFinished(outgroup);
		}
	}
	
	private ConditionTreeNode expandTree (NodePair pair) {
		ConditionTreeNode newNode = new ConditionTreeNode(null);
		
		// We now need to calculate the distances of the two daughter
		// nodes from the new node.  The distance from the new node is
		// simply this sum divided by the total nodes minus 2
		float r1 = distanceSum(pair.node1)/(nodes.size()-2);
		float r2 = distanceSum(pair.node2)/(nodes.size()-2);
		
		float d1 = (pair.distance + r1 - r2)/2;
		float d2 = pair.distance - d1;

		
		// Because we're constructing our trees from correlation data
		// and not from phylogenetic distances it is possible to get
		// negative branch lengths.  These don't have any real meaning
		// and make the trees difficult to interpret.  Apparently the
		// usual thing to do if this occurs is to set the distance to 0.
		// This seems kind of wrong, but but I've found a few sources
		// all of which suggest the same thing so who am I to argue?
		if (d1<0) d1=0;
		if (d2<0) d2=0;

		
		//		System.out.println("R1="+r1+" R2="+r2+" D1="+d1+" D2="+d2+" Nodes="+nodes.size()+" Pair Dist="+pair.distance);
		
		
		// We can now add the two closest nodes to the new node
		newNode.add(pair.node1,new Float(d1));
		newNode.add(pair.node2,new Float(d2));
		
		// We now need to calculate the distances to the new node
		// from all remaining nodes.
		Enumeration e = nodes.elements();
		while (e.hasMoreElements()) {
			ConditionTreeNode thisNode = (ConditionTreeNode)e.nextElement();
			if (thisNode == pair.node1 || thisNode == pair.node2) {
				continue;
			}
			
			float newDist = ((getDistance(pair.node1,thisNode)+getDistance(pair.node2,thisNode))-getDistance(pair.node1,pair.node2))/2;
	
			if (newDist < 0) {
				System.err.println("Negative distance dXm="+getDistance(pair.node1,thisNode)+" dYm="+getDistance(pair.node2,thisNode)+" dXY="+getDistance(pair.node1,pair.node2));
			}
			
			((Hashtable)matrix.get(thisNode)).put(newNode,new Float(newDist));
			
		}

		// We can now remove the closest pair from the list of available nodes
		nodes.removeElement(pair.node1);
		nodes.removeElement(pair.node2);

		
		// Finally add the new node to the list of available nodes
		nodes.add(newNode);
		matrix.put(newNode,new Hashtable());
		
		return newNode;
	}
	
	private float getDistance (ConditionTreeNode n1, ConditionTreeNode n2) {
		Enumeration e1 = nodes.elements();
		
		while (e1.hasMoreElements()) {
			ConditionTreeNode n = (ConditionTreeNode)e1.nextElement();
			if (n == n1) {
				return ((Float)((Hashtable)matrix.get(n1)).get(n2)).floatValue();
			}
			if (n == n2) {
				return ((Float)((Hashtable)matrix.get(n2)).get(n1)).floatValue();				
			}
		}	
		System.err.println("Didn't find "+n1+" or "+n2+" - giving default");
		return 3;
	}
	
	private float distanceSum (ConditionTreeNode n) {
		// Returns the sum of all distances for currently valid matrix nodes.

		float distanceTotal = 0;
		
		Enumeration e1 = nodes.elements();
	
		while (e1.hasMoreElements()) {
			ConditionTreeNode n1 = (ConditionTreeNode)e1.nextElement();
			if (n1 == n) {
				continue;
			}
			else {
				distanceTotal +=getDistance(n,n1);				
			}
		}		
		
		return distanceTotal;
	}
	
	private NodePair findClosestPair () {
		
		Enumeration e1 = nodes.elements();
		
		float smallestDist = -1;
		NodePair closestPair = null;
		
		while (e1.hasMoreElements()) {
			ConditionTreeNode n1 = (ConditionTreeNode)e1.nextElement();
			
			Enumeration e2 = nodes.elements();
			while (e2.hasMoreElements()) {
				if (e2.nextElement() == n1)
					break;
			}
			while (e2.hasMoreElements()) {
				ConditionTreeNode n2 = (ConditionTreeNode)e2.nextElement();
				float thisDist = ((Float)((Hashtable)matrix.get(n1)).get(n2)).floatValue();
				if (smallestDist < 0) {
					smallestDist = thisDist;
					closestPair = new NodePair(n1,n2,smallestDist);
				}
				else if (thisDist < smallestDist) {
					smallestDist = thisDist;
					closestPair = new NodePair(n1,n2,smallestDist);
				}
				
			}
		}
		
//		System.out.println("Closest pair are "+closestPair.node1+" and "+closestPair.node2+" with distance "+closestPair.distance);
		
		return closestPair;
	}
	
	private class NodePair {
		
		public ConditionTreeNode node1;
		public ConditionTreeNode node2;
		public float distance;
		
		private NodePair (ConditionTreeNode n1, ConditionTreeNode n2, float distance) {
			node1 = n1;
			node2 = n2;
			this.distance = distance;
		}
	}


}
